@@ -21,6 +21,8 @@
 
 #include "nanoarrow.h"
 
+#include "util.h"
+
 // Needed for the list_of materializer
 #include "convert.h"
 
@@ -31,6 +33,7 @@
 #include "materialize_dbl.h"
 #include "materialize_difftime.h"
 #include "materialize_int.h"
+#include "materialize_int64.h"
 #include "materialize_lgl.h"
 #include "materialize_posixct.h"
 #include "materialize_unspecified.h"
@@ -109,6 +112,16 @@ SEXP nanoarrow_materialize_realloc(SEXP ptype, R_xlen_t len) {
   SEXP result;
 
   if (Rf_isObject(ptype)) {
+    // There may be a more accurate test that more precisely captures the case
+    // where a user has specified a valid ptype that doesn't work in a preallocate
+    // + fill conversion.
+    if (Rf_inherits(ptype, "factor")) {
+      SEXP levels = Rf_getAttrib(ptype, R_LevelsSymbol);
+      if (Rf_length(levels) == 0) {
+        Rf_error("Can't allocate ptype of class 'factor' with empty levels");
+      }
+    }
+
     if (nanoarrow_ptype_is_data_frame(ptype)) {
       R_xlen_t num_cols = Rf_xlength(ptype);
       result = PROTECT(Rf_allocVector(VECSXP, num_cols));
@@ -148,6 +161,10 @@ static void fill_vec_with_nulls(SEXP x, R_xlen_t offset, R_xlen_t len) {
   }
 
   switch (TYPEOF(x)) {
+    case RAWSXP:
+      // Not perfect: raw() doesn't really support NA in R
+      memset(RAW(x), 0, len * sizeof(char));
+      break;
     case LGLSXP:
     case INTSXP: {
       int* values = INTEGER(x);
@@ -163,6 +180,17 @@ static void fill_vec_with_nulls(SEXP x, R_xlen_t offset, R_xlen_t len) {
       }
       return;
     }
+    case CPLXSXP: {
+      Rcomplex* values = COMPLEX(x);
+      Rcomplex na_value;
+      na_value.r = NA_REAL;
+      na_value.i = NA_REAL;
+
+      for (R_xlen_t i = 0; i < len; i++) {
+        values[offset + i] = na_value;
+      }
+      return;
+    }
     case STRSXP:
       for (R_xlen_t i = 0; i < len; i++) {
         SET_STRING_ELT(x, offset + i, NA_STRING);
@@ -178,12 +206,122 @@ static void fill_vec_with_nulls(SEXP x, R_xlen_t offset, R_xlen_t len) {
   }
 }
 
+static void copy_vec_into(SEXP x, SEXP dst, R_xlen_t offset, R_xlen_t len) {
+  if (nanoarrow_ptype_is_data_frame(dst)) {
+    if (!nanoarrow_ptype_is_data_frame(x)) {
+      Rf_error("Expected record-style vctr result but got non-record-style result");
+    }
+
+    R_xlen_t x_len = nanoarrow_data_frame_size(x);
+    if (len != x_len) {
+      Rf_error("Unexpected data.frame row count in copy_vec_into()");
+    }
+
+    // This does not currently consider column names (i.e., it blindly copies
+    // by index).
+    if (Rf_xlength(x) != Rf_xlength(dst)) {
+      Rf_error("Unexpected data.frame column count in copy_vec_into()");
+    }
+
+    for (R_xlen_t i = 0; i < Rf_xlength(x); i++) {
+      copy_vec_into(VECTOR_ELT(x, i), VECTOR_ELT(dst, i), offset, len);
+    }
+
+    return;
+  } else if (nanoarrow_ptype_is_data_frame(x)) {
+    Rf_error("Expected non-record-style vctr result but got record-style result");
+  }
+
+  if (TYPEOF(dst) != TYPEOF(x)) {
+    Rf_error("Unexpected SEXP type in result copy_vec_into()");
+  }
+
+  if (Rf_length(x) != len) {
+    Rf_error("Unexpected length of result in copy_vec_into()");
+  }
+
+  switch (TYPEOF(dst)) {
+    case RAWSXP:
+      memcpy(RAW(dst) + offset, RAW(x), len * sizeof(uint8_t));
+      break;
+    case REALSXP:
+      memcpy(REAL(dst) + offset, REAL(x), len * sizeof(double));
+      break;
+    case INTSXP:
+    case LGLSXP:
+      memcpy(INTEGER(dst) + offset, INTEGER(x), len * sizeof(int));
+      break;
+    case CPLXSXP:
+      memcpy(COMPLEX(dst) + offset, COMPLEX(x), len * sizeof(Rcomplex));
+      break;
+    case STRSXP:
+      for (R_xlen_t i = 0; i < len; i++) {
+        SET_STRING_ELT(dst, offset + i, STRING_ELT(x, i));
+      }
+      break;
+    case VECSXP:
+      for (R_xlen_t i = 0; i < len; i++) {
+        SET_VECTOR_ELT(dst, offset + i, VECTOR_ELT(x, i));
+      }
+      break;
+    default:
+      Rf_error("Unhandled SEXP type in copy_vec_into()");
+      break;
+  }
+}
+
+static int nanoarrow_materialize_other(struct RConverter* converter,
+                                       SEXP converter_xptr) {
+  // Ensure that we have a ptype SEXP to send in the call back to R
+  if (converter->ptype_view.ptype == R_NilValue) {
+    SEXP ptype = PROTECT(nanoarrow_alloc_type(converter->ptype_view.vector_type, 0));
+    converter->ptype_view.ptype = ptype;
+    SET_VECTOR_ELT(R_ExternalPtrProtected(converter_xptr), 0, ptype);
+    UNPROTECT(1);
+  }
+
+  // A unique situation where we don't want owning external pointers because we know
+  // these are protected for the duration of our call into R and because we don't want
+  // the underlying array to be released and invalidate the converter. The R code in
+  // convert_fallback_other() takes care of ensuring an independent copy with the correct
+  // offset/length.
+  SEXP schema_xptr =
+      PROTECT(R_MakeExternalPtr(converter->schema_view.schema, R_NilValue, R_NilValue));
+  Rf_setAttrib(schema_xptr, R_ClassSymbol, nanoarrow_cls_schema);
+  // We do need to set the protected member of the array external pointer to signal that
+  // it is not an independent array (i.e., force a shallow copy).
+  SEXP array_xptr = PROTECT(
+      R_MakeExternalPtr(converter->array_view.array, schema_xptr, converter_xptr));
+  Rf_setAttrib(array_xptr, R_ClassSymbol, nanoarrow_cls_array);
+
+  SEXP offset_sexp =
+      PROTECT(Rf_ScalarReal(converter->src.array_view->offset + converter->src.offset));
+  SEXP length_sexp = PROTECT(Rf_ScalarReal(converter->src.length));
+
+  SEXP fun = PROTECT(Rf_install("convert_fallback_other"));
+  SEXP call = PROTECT(
+      Rf_lang5(fun, array_xptr, offset_sexp, length_sexp, converter->ptype_view.ptype));
+  SEXP result_src = PROTECT(Rf_eval(call, nanoarrow_ns_pkg));
+
+  // Copy the result into a slice of dst
+  copy_vec_into(result_src, converter->dst.vec_sexp, converter->dst.offset,
+                converter->dst.length);
+
+  UNPROTECT(7);
+  return NANOARROW_OK;
+}
+
 static int nanoarrow_materialize_data_frame(struct RConverter* converter,
                                             SEXP converter_xptr) {
   if (converter->ptype_view.vector_type != VECTOR_TYPE_DATA_FRAME) {
     return EINVAL;
   }
 
+  // Make sure we error for dictionary types
+  if (converter->src.array_view->array->dictionary != NULL) {
+    return EINVAL;
+  }
+
   SEXP converter_shelter = R_ExternalPtrProtected(converter_xptr);
   SEXP child_converter_xptrs = VECTOR_ELT(converter_shelter, 3);
 
@@ -256,6 +394,11 @@ static int nanoarrow_materialize_list_of(struct RConverter* converter,
   struct ArrayViewSlice* src = &converter->src;
   struct VectorSlice* dst = &converter->dst;
 
+  // Make sure we error for dictionary types
+  if (src->array_view->array->dictionary != NULL) {
+    return EINVAL;
+  }
+
   const int32_t* offsets = src->array_view->buffer_views[1].data.as_int32;
   const int64_t* large_offsets = src->array_view->buffer_views[1].data.as_int64;
   int64_t raw_src_offset = src->array_view->array->offset + src->offset;
@@ -309,11 +452,16 @@ static int nanoarrow_materialize_list_of(struct RConverter* converter,
   return NANOARROW_OK;
 }
 
-int nanoarrow_materialize(struct RConverter* converter, SEXP converter_xptr) {
+static int nanoarrow_materialize_base(struct RConverter* converter, SEXP converter_xptr) {
   struct ArrayViewSlice* src = &converter->src;
   struct VectorSlice* dst = &converter->dst;
   struct MaterializeOptions* options = converter->options;
 
+  // Make sure extension conversion calls into R
+  if (converter->schema_view.extension_name.size_bytes > 0) {
+    return nanoarrow_materialize_other(converter, converter_xptr);
+  }
+
   switch (converter->ptype_view.vector_type) {
     case VECTOR_TYPE_UNSPECIFIED:
       return nanoarrow_materialize_unspecified(src, dst, options);
@@ -324,13 +472,15 @@ int nanoarrow_materialize(struct RConverter* converter, SEXP converter_xptr) {
     case VECTOR_TYPE_DBL:
       return nanoarrow_materialize_dbl(converter);
     case VECTOR_TYPE_CHR:
-      return nanoarrow_materialize_chr(src, dst, options);
+      return nanoarrow_materialize_chr(converter);
     case VECTOR_TYPE_POSIXCT:
       return nanoarrow_materialize_posixct(converter);
     case VECTOR_TYPE_DATE:
       return nanoarrow_materialize_date(converter);
     case VECTOR_TYPE_DIFFTIME:
       return nanoarrow_materialize_difftime(converter);
+    case VECTOR_TYPE_INTEGER64:
+      return nanoarrow_materialize_int64(src, dst, options);
     case VECTOR_TYPE_BLOB:
       return nanoarrow_materialize_blob(src, dst, options);
     case VECTOR_TYPE_LIST_OF:
@@ -338,6 +488,16 @@ int nanoarrow_materialize(struct RConverter* converter, SEXP converter_xptr) {
     case VECTOR_TYPE_DATA_FRAME:
       return nanoarrow_materialize_data_frame(converter, converter_xptr);
     default:
-      return ENOTSUP;
+      return nanoarrow_materialize_other(converter, converter_xptr);
+  }
+}
+
+int nanoarrow_materialize(struct RConverter* converter, SEXP converter_xptr) {
+  int result = nanoarrow_materialize_base(converter, converter_xptr);
+
+  if (result != NANOARROW_OK) {
+    return nanoarrow_materialize_other(converter, converter_xptr);
+  } else {
+    return NANOARROW_OK;
   }
 }
